// created: 10.8.01 modified: 8.8.01 final clean-up: 24.4.02
// Stephan Dale
// NN user library

#include <stdlib.h>
#include "maverik.h"

#include "netlib.h"
#include "sjd_backprop.h" // for the BPNN datatype
#include "nettools.h" // for the entry datatype

/**************************************************************************************************/

/***********/
/* globals */
/***********/
float shoulderWidth, upperArmLength, lowerArmLength;

/**************************************************************************************************/

/********************************/
/* external function prototypes */
/********************************/
// from nettools
extern entry *load_single_netinput(BPNN *net, float *data);
extern void set_netindata(entry *pattern, BPNN *net);
// from avatarlib
extern getPY(MAV_vector vec, float *pitch, float *yaw);

/********************************/
/* internal function prototypes */
/********************************/
void preProcess(MAV_vector hand, float handRoll, float handPitch, float handYaw, float *x, float *y, float *z, float *roll, float *pitch, float *yaw);
void postProcess(float yaw, float pitch, float *armYaw, float *armPitch);

/**************************************************************************************************/

/*****************************/
/* pre-process network input */
/*****************************/
void preProcess(MAV_vector hand, float handRoll, float handPitch, float handYaw, float *x, float *y, float *z, float *roll, float *pitch, float *yaw)
 {
 float tmp, maxMag;
  // normalise hand vector
 maxMag = shoulderWidth + upperArmLength + lowerArmLength; // calculate max magnitude (using max possible arm extension)
 tmp = mav_vectorMag(hand);
 hand = mav_vectorScalar(hand, (tmp/maxMag)/tmp);
 // map onto the range 0 -> 1 (from range -1 -> 1)
 // note: must remove any out of range (from innaccurate user proportions)?
 hand.x = (hand.x/2.0)+0.5;
 hand.y = (hand.y/2.0)+0.5;
 hand.z = (hand.z/2.0)+0.5;

 // set inputs
 *x = hand.x;
 *y = hand.y;
 *z = hand.z;

 // normalise angles
 if (handRoll<0) handRoll += 360; // ensure >=0
 if (handPitch<0) handPitch += 360;
 if (handYaw<0) handYaw += 360;
 handRoll = handRoll/360.0; // max roll is 360 (->1)
 handPitch = handPitch/360.0;
 handYaw = handYaw/360.0;

 // set inputs
 *roll = handRoll;
 *pitch = handPitch;
 *yaw = handYaw;
 }

/*******************************/
/* post-process network output */
/*******************************/
void postProcess(float yaw, float pitch, float *armYaw, float *armPitch)
 {
 MAV_vector vec;

 // convert to degrees (from range 0-1 to range 0-360)
 yaw = yaw*360.0;
 pitch = pitch*360.0;

 // calculate vector relative to back
 vec = mav_vectorRotate(mav_vectorScalar(MAV_Y_VECTOR,-1), mav_vectorScalar(MAV_Z_VECTOR, -1), pitch); // pitch vector (in xy plane as z==0)
 vec = mav_vectorRotate(vec, MAV_Y_VECTOR, yaw); // yaw vector (from xy plane into xyz space)

 // shift so relative to shoulder
 vec.x += shoulderWidth;

 // get pitch and yaw of arm relative to shoulder
 getPY(vec, &pitch, &yaw);

 *armYaw = yaw;
 *armPitch = pitch;
 }

/*************************************/
/* get network output given all data */
/*************************************/
int predict(BPNN *net, float sWidth, float uALength, float lALength, MAV_vector hand, float handRoll, float handPitch, float handYaw, float *armYaw, float *armPitch)
 {
 entry *pattern;
 float indata[7];
 float *outdata;

 if (net->input_n!=6 || net->output_n!=2)
   return 0;

 shoulderWidth = sWidth;
 upperArmLength = uALength;
 lowerArmLength = lALength;

 // pre-process
 preProcess(hand, handRoll, handPitch, handYaw, &indata[1], &indata[2], &indata[3], &indata[4], &indata[5], &indata[6]);

 pattern = load_single_netinput(net,indata); // set input pattern
 set_netindata(pattern,net); // load pattern
 outdata = sjd_bpnn_predict(net);//, 0.3, 0.3);  // get output

 // post-process (convert so relative to shoulder rather than back)
 postProcess(outdata[1], outdata[2], armYaw, armPitch);

 return 1;
 }

/**************************************************/
/* get network output given data without hand RPY */
/**************************************************/
/*int predictNoHandRPY(BPNN *net, MAV_vector hand, float *pitch, float *yaw)
 {
 entry *pattern;
 float indata[4];
 float *outdata;

 if (net->input_n!=3 || net->output_n!=2)
   return 0;

 indata[1] = (float) hand.x;
 indata[2] = (float) hand.y;
 indata[3] = (float) hand.z;

 pattern = load_single_netinput(net,indata); // set input pattern
 set_netindata(pattern,net); // load pattern
 outdata = sjd_bpnn_predict(net);//, 0.3, 0.3);  // get output

 *pitch = (float) outdata[1];
 *yaw = (float) outdata[2];

 return 1;
 } */

/*****************************/
/* wrapper for sjd_bpnn_read */
/*****************************/
BPNN *loadNet(char *filename)
 {
 return sjd_bpnn_read(filename);
 }

/*****************************/
/* wrapper for sjd_bpnn_free */
/*****************************/
void freeNet(BPNN *net)
 {
 sjd_bpnn_free(net);
 }

/**************************************************************************************************/
